use enum_ordinalize::Ordinalize;
use enumset::{enum_set, EnumSet, EnumSetType};

use crate::util::Endianness;

#[derive(Debug, Ordinalize, EnumSetType)]
#[repr(u8)]
pub enum LoupedeckKnob {
    LeftTop = 0x01,
    LeftMiddle = 0x02,
    LeftBottom = 0x03,
    RightTop = 0x04,
    RightMiddle = 0x05,
    RightBottom = 0x06,
}

impl LoupedeckKnob {
    fn is_left(&self) -> bool {
        matches!(self, LoupedeckKnob::LeftTop | LoupedeckKnob::LeftMiddle | LoupedeckKnob::LeftBottom)
    }

    fn row(&self) -> u8 {
        match self {
            LoupedeckKnob::LeftTop | LoupedeckKnob::RightTop => 0,
            LoupedeckKnob::LeftMiddle | LoupedeckKnob::RightMiddle => 1,
            LoupedeckKnob::LeftBottom | LoupedeckKnob::RightBottom => 2,
        }
    }
}

#[derive(Debug, Ordinalize, EnumSetType)]
#[repr(u8)]
pub enum LoupedeckButton {
    N0 = 0x07,
    N1 = 0x08,
    N2 = 0x09,
    N3 = 0x0a,
    N4 = 0x0b,
    N5 = 0x0c,
    N6 = 0x0d,
    N7 = 0x0e,
}

#[derive(Debug)]
#[non_exhaustive]
pub struct LoupedeckDeviceDisplayConfiguration {
    pub id: u8,
    pub dpi: f32,
    pub width: u16,
    pub height: u16,
    pub local_offset_x: u16,
    pub local_offset_y: u16,
    pub global_offset_x: u16,
    pub global_offset_y: u16,
    pub endianness: Endianness,
}

impl Eq for LoupedeckDeviceDisplayConfiguration {
    fn assert_receiver_is_total_eq(&self) {}
}

impl PartialEq for LoupedeckDeviceDisplayConfiguration {
    fn eq(&self, other: &Self) -> bool {
        std::ptr::eq(self, other)
    }
}

#[derive(Debug)]
#[non_exhaustive]
pub struct LoupedeckDeviceKeyGridCharacteristics {
    pub rows: u8,
    pub columns: u8,
    pub display: LoupedeckDeviceDisplayConfiguration,
}

#[derive(Debug)]
pub struct LoupedeckDisplayRect {
    pub x: u16,
    pub y: u16,
    pub w: u16,
    pub h: u16,
}

impl LoupedeckDeviceKeyGridCharacteristics {
    pub fn key_size(&self) -> (u16, u16) {
        // Assuming the sizes are integers
        (self.display.width / self.columns as u16, self.display.height / self.rows as u16)
    }

    pub fn get_key_at_local_coordinates(&self, x: u16, y: u16) -> Option<u8> {
        let (column_width, row_height) = self.key_size();

        if x >= self.display.width || y >= self.display.height {
            return None;
        }

        let column = (x / column_width) as u8;
        let row = (y / row_height) as u8;

        Some(row * self.columns + column)
    }

    pub fn get_key_at_global_coordinates(&self, x: u16, y: u16) -> Option<u8> {
        if x < self.display.global_offset_x || y < self.display.global_offset_y {
            return None;
        }

        let local_x = x - self.display.global_offset_x;
        let local_y = y - self.display.global_offset_y;

        self.get_key_at_local_coordinates(local_x, local_y)
    }

    pub fn get_local_key_rect(&self, key_index: u8) -> Option<LoupedeckDisplayRect> {
        if key_index >= self.rows * self.columns {
            return None;
        }

        let (column_width, row_height) = self.key_size();
        let row = (key_index / self.columns) as u16;
        let column = (key_index % self.columns) as u16;

        Some(LoupedeckDisplayRect {
            x: column * column_width,
            y: row * row_height,
            w: column_width,
            h: row_height,
        })
    }
}

#[derive(Debug)]
#[non_exhaustive]
pub struct LoupedeckDeviceCharacteristics {
    pub vendor_id: u16,
    pub product_id: u16,
    pub name: &'static str,
    pub available_knobs: EnumSet<LoupedeckKnob>,
    pub available_buttons: EnumSet<LoupedeckButton>,
    pub key_grid: LoupedeckDeviceKeyGridCharacteristics,
    pub knob_displays: Option<(LoupedeckDeviceDisplayConfiguration, LoupedeckDeviceDisplayConfiguration)>,
}

impl Eq for LoupedeckDeviceCharacteristics {
    fn assert_receiver_is_total_eq(&self) {}
}

impl PartialEq for LoupedeckDeviceCharacteristics {
    fn eq(&self, other: &Self) -> bool {
        std::ptr::eq(self, other)
    }
}

impl LoupedeckDeviceCharacteristics {
    pub fn knob_rows(&self) -> u8 {
        self.available_knobs.iter().map(|k| k.row()).max().map(|v| v + 1).unwrap_or(0)
    }

    pub fn get_display_at_coordinates(&self, x: u16, y: u16) -> Option<&LoupedeckDeviceDisplayConfiguration> {
        let check = |display: &LoupedeckDeviceDisplayConfiguration| {
            x >= display.global_offset_x
                && x <= display.global_offset_x + display.width
                && y >= display.global_offset_y
                && y <= display.global_offset_y + display.height
        };

        if check(&self.key_grid.display) {
            Some(&self.key_grid.display)
        } else if let Some(knob_displays) = &self.knob_displays {
            if check(&knob_displays.0) {
                Some(&knob_displays.0)
            } else if check(&knob_displays.1) {
                Some(&knob_displays.1)
            } else {
                None
            }
        } else {
            None
        }
    }

    pub fn get_display_and_rect_for_knob(&self, knob: LoupedeckKnob) -> Option<(&LoupedeckDeviceDisplayConfiguration, LoupedeckDisplayRect)> {
        if !self.available_knobs.contains(knob) {
            return None;
        }

        if let Some(knob_displays) = &self.knob_displays {
            let display = if knob.is_left() { &knob_displays.0 } else { &knob_displays.1 };
            let row = knob.row() as u16;
            let rows = self.knob_rows() as u16;

            let row_height = display.height / rows;

            let rect = LoupedeckDisplayRect {
                x: 0,
                y: row_height * row,
                w: display.width,
                h: row_height,
            };

            Some((display, rect))
        } else {
            None
        }
    }
}

static LOUPEDECK_LIVE_CHARACTERISTIC: LoupedeckDeviceCharacteristics = LoupedeckDeviceCharacteristics {
    vendor_id: 0x2ec2,
    product_id: 0x0004,
    name: "Loupedeck Live",
    available_knobs: enum_set!(
        LoupedeckKnob::LeftTop
            | LoupedeckKnob::LeftMiddle
            | LoupedeckKnob::LeftBottom
            | LoupedeckKnob::RightTop
            | LoupedeckKnob::RightMiddle
            | LoupedeckKnob::RightBottom
    ),
    available_buttons: enum_set!(
        LoupedeckButton::N0
            | LoupedeckButton::N1
            | LoupedeckButton::N2
            | LoupedeckButton::N3
            | LoupedeckButton::N4
            | LoupedeckButton::N5
            | LoupedeckButton::N6
            | LoupedeckButton::N7
    ),
    key_grid: LoupedeckDeviceKeyGridCharacteristics {
        rows: 3,
        columns: 4,
        display: LoupedeckDeviceDisplayConfiguration {
            id: 0x4d,
            dpi: 142.875,
            width: 360,
            height: 270,
            local_offset_x: 60,
            local_offset_y: 0,
            global_offset_x: 60,
            global_offset_y: 0,
            endianness: Endianness::LittleEndian,
        },
    },
    knob_displays: Some((
        LoupedeckDeviceDisplayConfiguration {
            id: 0x4d,
            dpi: 142.875,
            width: 60,
            height: 270,
            local_offset_x: 0,
            local_offset_y: 0,
            global_offset_x: 0,
            global_offset_y: 0,
            endianness: Endianness::LittleEndian,
        },
        LoupedeckDeviceDisplayConfiguration {
            id: 0x4d,
            dpi: 142.875,
            width: 60,
            height: 270,
            local_offset_x: 420,
            local_offset_y: 0,
            global_offset_x: 420,
            global_offset_y: 0,
            endianness: Endianness::LittleEndian,
        },
    )),
};

pub static CHARACTERISTICS: [&LoupedeckDeviceCharacteristics; 1] = [&LOUPEDECK_LIVE_CHARACTERISTIC];
