use std::sync::Arc;

use parse_display::Display;
use pulseaudio_volume_interface::PulseAudioVolumeInterface;
use regex::Regex;
use serde::Deserialize;
use tokio::sync::broadcast;

use deckster_mode::shared::handler_communication::{
    HandlerCommand, HandlerEvent, HandlerInitializationError, InitialHandlerMessage, KnobEvent, RotationDirection,
};
use deckster_mode::shared::path::KnobPath;
use deckster_mode::shared::state::KnobStyleByStateMap;
use deckster_mode::{send_command, DecksterHandler};
use pulseaudio_volume_interface::{EntityKind as PaEntityKind, EntityMetadata as PaEntityMetadata, EntityState as PaEntityState};

#[derive(Debug, Clone, Deserialize)]
pub struct KnobConfig {
    pub target: Target,
    pub delta: Option<f32>,
    #[serde(default)]
    pub disable_press_to_mute: bool,
    #[serde(default)]
    pub disable_press_to_unmute: bool,
    #[serde(default)]
    pub muted_turn_action: MutedTurnAction,
    #[serde(default)]
    pub style: KnobStyleByStateMap<State>,
}

#[derive(Debug, Clone, Eq, PartialEq, Deserialize, Display)]
#[serde(rename_all = "kebab-case")]
#[display(style = "kebab-case")]
pub enum TargetKind {
    Input,
    Output,
    Application,
}

#[derive(Debug, Clone, Deserialize)]
pub struct Target {
    #[serde(rename = "type")]
    kind: TargetKind,
    predicates: Vec<TargetPredicate>,
}

#[derive(Debug, Clone, Deserialize)]
pub struct TargetPredicate {
    property: TargetPredicateProperty,
    #[serde(flatten)]
    pattern: TargetPredicatePattern,
}

#[derive(Debug, Clone, Deserialize)]
#[serde(untagged)]
pub enum TargetPredicatePattern {
    Static {
        value: String,
    },
    Regex {
        #[serde(with = "serde_regex")]
        regex: Regex,
    },
}

#[derive(Debug, Clone, Eq, PartialEq, Deserialize, Display)]
#[serde(rename_all = "kebab-case")]
#[display(style = "kebab-case")]
pub enum TargetPredicateProperty {
    Description,
    InternalName,
    ApplicationName,
    BinaryName,
}

#[derive(Debug, Clone, Default, Eq, PartialEq, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum MutedTurnAction {
    #[default]
    Ignore,
    Normal,
    UnmuteAtZero,
    Unmute,
}

#[derive(Debug, Clone, Default, Eq, PartialEq, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum Direction {
    #[default]
    Output,
    Input,
}

#[derive(Debug, Clone, Eq, PartialEq, Hash, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum State {
    Inactive,
    Active,
    Muted,
}

fn get_volume_from_cv(channel_volumes: &[f32]) -> f32 {
    *channel_volumes.iter().max_by(|a, b| a.partial_cmp(b).unwrap()).unwrap()
}

fn state_matches(target: &Target, state: &PaEntityState) -> bool {
    if !match target.kind {
        TargetKind::Input => state.kind() == PaEntityKind::Source,
        TargetKind::Output => state.kind() == PaEntityKind::Sink,
        TargetKind::Application => state.kind() == PaEntityKind::SinkInput,
    } {
        return false;
    }

    static EMPTY_STRING: String = String::new();

    target.predicates.iter().all(|p| {
        let v = match (&p.property, state.metadata()) {
            (TargetPredicateProperty::InternalName, PaEntityMetadata::Sink { name, .. }) => Some(name),
            (TargetPredicateProperty::InternalName, PaEntityMetadata::Source { name, .. }) => Some(name),
            (TargetPredicateProperty::InternalName, PaEntityMetadata::SinkInput { .. }) => None,
            (TargetPredicateProperty::Description, PaEntityMetadata::Sink { description, .. }) => Some(description),
            (TargetPredicateProperty::Description, PaEntityMetadata::Source { description, .. }) => Some(description),
            (TargetPredicateProperty::Description, PaEntityMetadata::SinkInput { description, .. }) => Some(description),
            (TargetPredicateProperty::ApplicationName, PaEntityMetadata::Sink { .. }) => None,
            (TargetPredicateProperty::ApplicationName, PaEntityMetadata::Source { .. }) => None,
            (TargetPredicateProperty::ApplicationName, PaEntityMetadata::SinkInput { application_name, .. }) => {
                Some(application_name.as_ref().unwrap_or(&EMPTY_STRING))
            }
            (TargetPredicateProperty::BinaryName, PaEntityMetadata::Sink { .. }) => None,
            (TargetPredicateProperty::BinaryName, PaEntityMetadata::Source { .. }) => None,
            (TargetPredicateProperty::BinaryName, PaEntityMetadata::SinkInput { binary_name, .. }) => Some(binary_name.as_ref().unwrap_or(&EMPTY_STRING)),
        };

        if let Some(v) = v {
            match &p.pattern {
                TargetPredicatePattern::Static { value } => value == v,
                TargetPredicatePattern::Regex { regex } => regex.is_match(v),
            }
        } else {
            log::warn!("Property \"{}\" is not available for targets of type \"{}\"", &p.property, &target.kind);

            false
        }
    })
}

async fn manage_knob(path: KnobPath, config: KnobConfig, mut events: broadcast::Receiver<(KnobPath, KnobEvent)>, pa_volume_interface: PulseAudioVolumeInterface) {
    let (initial_state, mut volume_states) = pa_volume_interface.subscribe_to_state();

    let mut entity_state: Option<Arc<PaEntityState>> = initial_state
        .entities_by_id()
        .values()
        .find(|entity| state_matches(&config.target, entity))
        .map(Arc::clone);

    let update_knob_value = {
        let config = &config;
        let path = path.clone();

        move |entity_state: &Option<Arc<PaEntityState>>| {
            send_command(HandlerCommand::SetKnobValue {
                path: path.clone(),
                value: entity_state.as_ref().map(|s| {
                    if s.is_muted() && config.muted_turn_action == MutedTurnAction::UnmuteAtZero {
                        0.0
                    } else {
                        get_volume_from_cv(&s.channel_volumes())
                    }
                }),
            });
        }
    };

    let update_knob_style = {
        let config = &config;
        let path = path.clone();

        move |entity_state: &Option<Arc<PaEntityState>>| {
            let state = match entity_state {
                None => State::Inactive,
                Some(s) if s.is_muted() => State::Muted,
                Some(_) => State::Active,
            };

            let mut style = config.style.get(&state).cloned();

            if let Some(ref mut s) = &mut style {
                let v = entity_state.as_ref().map(|s| get_volume_from_cv(&s.channel_volumes()));

                if let Some(ref mut label) = &mut s.label {
                    if let Some(v) = v {
                        // v is only None when state is State::Inactive
                        *label = label.replace("{percentage}", &((v * 100.0).round() as u32).to_string());
                    }
                }
            }

            send_command(HandlerCommand::SetKnobStyle {
                path: path.clone(),
                value: style,
            });
        }
    };

    update_knob_style(&entity_state);
    update_knob_value(&entity_state);

    loop {
        tokio::select! {
            Ok(volume_state) = volume_states.recv() => {
                let new_entity_state = volume_state.entities_by_id().values().find(|entity| state_matches(&config.target, entity)).map(Arc::clone);
                if entity_state != new_entity_state {
                    entity_state = new_entity_state;
                    update_knob_style(&entity_state);
                    update_knob_value(&entity_state);
                }
            }

            Ok((event_path, event)) = events.recv() => {
                if event_path != path {
                    continue;
                }

                if let Some(entity_state) = &entity_state {
                    match event {
                        KnobEvent::Rotate { direction } => {
                            let factor: f32 = match direction {
                                RotationDirection::Clockwise => 1.0,
                                RotationDirection::Counterclockwise => -1.0,
                            };

                            let mut current_v = get_volume_from_cv(&entity_state.channel_volumes());

                            if entity_state.is_muted() {
                                match config.muted_turn_action {
                                    MutedTurnAction::Ignore => continue,
                                    MutedTurnAction::UnmuteAtZero => {
                                        current_v = 0.0
                                    }
                                    MutedTurnAction::Unmute => {},
                                    MutedTurnAction::Normal => {}
                                }
                            }

                            let new_v = (current_v + (factor * config.delta.unwrap_or(0.01))).clamp(0.0, 1.0);
                            if new_v > 0.0 && matches!(config.muted_turn_action, MutedTurnAction::Unmute | MutedTurnAction::UnmuteAtZero) {
                                pa_volume_interface.set_is_muted(*entity_state.id(), false);
                            }

                            pa_volume_interface.set_channel_volumes(*entity_state.id(), vec![new_v; entity_state.channel_volumes().len()]);
                        }
                        KnobEvent::Press => {
                            let is_muted = entity_state.is_muted();

                            if (is_muted && !config.disable_press_to_unmute) || (!is_muted && !config.disable_press_to_mute) {
                                pa_volume_interface.set_is_muted(*entity_state.id(), !is_muted)
                            }
                        }
                        _ => {}
                    }
                }
            }
        }
    }
}

pub struct Handler {
    events_sender: broadcast::Sender<(KnobPath, KnobEvent)>,
    #[allow(unused)]
    runtime: tokio::runtime::Runtime,
}

impl Handler {
    pub fn new(data: InitialHandlerMessage<(), (), KnobConfig>) -> Result<Self, HandlerInitializationError> {
        let (events_sender, _) = broadcast::channel::<(KnobPath, KnobEvent)>(5);
        let pa_volume_interface = PulseAudioVolumeInterface::new("deckster handler".to_owned());

        let runtime = tokio::runtime::Builder::new_multi_thread().worker_threads(1).build().unwrap();

        for (path, config) in data.knob_configs {
            let events_receiver = events_sender.subscribe();
            runtime.spawn(manage_knob(path, config, events_receiver, pa_volume_interface.clone()));
        }

        Ok(Handler { events_sender, runtime })
    }
}

impl DecksterHandler for Handler {
    fn handle(&mut self, event: HandlerEvent) {
        if let HandlerEvent::Knob { path, event } = event {
            self.events_sender.send((path, event)).unwrap();
        }
    }
}
