use std::time::{Duration, Instant};
use tokio::sync::mpsc;
use tokio::sync::mpsc::error::TrySendError;
use tokio::sync::mpsc::{Receiver, Sender};
use tokio::time::timeout;

/// Sends a message into the output channel after a message in the input channel was received, with a delay of `duration`.
/// The delay is reset when a new message is reset.
pub fn spawn_debouncer(duration: Duration) -> (Sender<()>, Receiver<()>) {
    let (input_sender, mut input_receiver) = mpsc::channel::<()>(1);
    let (output_sender, output_receiver) = mpsc::channel::<()>(1);

    tokio::spawn(async move {
        'outer: loop {
            if input_receiver.recv().await.is_none() {
                break 'outer;
            }

            'inner: loop {
                match timeout(duration, input_receiver.recv()).await {
                    Ok(None) => break 'outer,
                    Ok(Some(_)) => continue 'inner,
                    Err(_) => {
                        if let Err(TrySendError::Closed(_)) = output_sender.try_send(()) {
                            break 'outer;
                        } else {
                            break 'inner;
                        }
                    }
                }
            }
        }
    });

    (input_sender, output_receiver)
}

pub fn format_duration(duration: Duration) -> String {
    let full_seconds = duration.as_secs();
    let full_minutes = full_seconds / 60;
    let hours = full_minutes / 60;
    let minutes = full_minutes % 60;
    let seconds = full_seconds % 60;

    if hours == 0 {
        format!("{:0>2}:{:0>2}", minutes, seconds)
    } else {
        format!("{:0>2}:{:0>2}:{:0>2}", hours, minutes, seconds)
    }
}

pub fn get_far_future() -> Instant {
    Instant::now() + Duration::from_secs(60 * 60 * 24 * 365 * 30) // 30 years
}
