use std::cell::RefCell;
use std::ops::Add;
use std::path::Path;
use std::sync::Arc;

use enum_ordinalize::Ordinalize;
use resvg::usvg::tiny_skia_path::IntSize;
use rgb::RGB8;
use tokio::sync::broadcast;

use deckster_shared::handler_communication::{HandlerCommand, HandlerEvent, KeyEvent, KeyTouchEventKind, KnobEvent, VibrationPattern};
use deckster_shared::path::{KeyPath, KeyPosition, KnobPath, KnobPosition};
use deckster_shared::state::{Key, Knob};
use loupedeck_serial::characteristics::{LoupedeckButton, LoupedeckDeviceKeyGridCharacteristics, LoupedeckDisplayRect, LoupedeckKnob};
use loupedeck_serial::commands::VibrationPattern as LSVibrationPattern;
use loupedeck_serial::device::LoupedeckDevice;
use loupedeck_serial::events::{LoupedeckEvent, RotationDirection};

use crate::coordinator::graphics::labels::LabelRenderer;
use crate::coordinator::graphics::{render_key, render_knob, GraphicsContext};
use crate::coordinator::state::State;
use crate::icons::IconManager;
use crate::model::coordinator_config::Config;
use crate::model::position::ButtonPosition;

#[derive(Debug, PartialEq, Clone)]
#[allow(clippy::enum_variant_names)]
pub enum CoordinatorCommand {
    SetActivePages { key_page_id: String, knob_page_id: String },
    SetRemoteHostIsActive { host_id: Box<str>, is_active: bool },
    SetAllRemoteHostsInactive,
}

enum IoWork {
    DeviceEvent(LoupedeckEvent),
    HandlerCommand(HandlerCommand),
    CoordinatorCommand(CoordinatorCommand),
}

pub struct IoWorkerContext {
    config: Arc<Config>,
    device: LoupedeckDevice,
    coordinator_commands_sender: flume::Sender<CoordinatorCommand>,
    events_sender: broadcast::Sender<HandlerEvent>,
    graphics: GraphicsContext,
    state: State,
}

impl IoWorkerContext {
    pub fn create(
        config_directory: &Path,
        config: Arc<Config>,
        device: LoupedeckDevice,
        coordinator_commands_sender: flume::Sender<CoordinatorCommand>,
        events_sender: broadcast::Sender<HandlerEvent>,
    ) -> Self {
        let buffer_endianness = device.characteristics().key_grid.display.endianness;
        let label_renderer = RefCell::new(LabelRenderer::new(config.label_font_family.as_ref()));
        let dpi = device.characteristics().key_grid.display.dpi;
        let icon_packs = Arc::clone(&config.icon_packs);
        let state = State::create(&config);

        IoWorkerContext {
            config,
            device,
            coordinator_commands_sender,
            events_sender,
            graphics: GraphicsContext {
                buffer_endianness,
                label_renderer,
                icon_manager: IconManager::new(config_directory.to_path_buf(), icon_packs, dpi),
            },
            state,
        }
    }
}

pub fn do_io_work(
    mut context: IoWorkerContext,
    coordinator_commands_receiver: &flume::Receiver<CoordinatorCommand>,
    handler_commands_receiver: &flume::Receiver<HandlerCommand>,
) {
    let device_events_receiver = context.device.events();

    loop {
        let work = flume::Selector::new()
            .recv(&device_events_receiver, |e| {
                IoWork::DeviceEvent(e.expect("the device events channel is not closed by the other side"))
            })
            .recv(coordinator_commands_receiver, |c| IoWork::CoordinatorCommand(c.unwrap()))
            .recv(handler_commands_receiver, |c| IoWork::HandlerCommand(c.unwrap()))
            .wait();

        match work {
            IoWork::DeviceEvent(event) => {
                if !handle_event(&mut context, event) {
                    break;
                }
            }
            IoWork::CoordinatorCommand(command) => handle_coordinator_command(&mut context, command),
            IoWork::HandlerCommand(command) => handle_handler_command(&mut context, command),
        }
    }
}

fn handle_event(context: &mut IoWorkerContext, event: LoupedeckEvent) -> bool {
    log::trace!("Handling event: {:?}", &event);

    let send_key_event = |path: KeyPath, event: KeyEvent| {
        log::trace!("Sending key event ({}): {:?}", &path, &event);
        context.events_sender.send(HandlerEvent::Key { path, event }).unwrap();
    };

    let send_knob_event = |path: KnobPath, event: KnobEvent| {
        log::trace!("Sending knob event ({:?}): {:?}", &path, &event);
        context.events_sender.send(HandlerEvent::Knob { path, event }).unwrap();
    };

    match event {
        LoupedeckEvent::Disconnected => return false,
        LoupedeckEvent::ButtonDown { button } => {
            let position = ButtonPosition::of(&button);
            let button_config = &context.config.buttons[position];

            if button == LoupedeckButton::N0 {
                context.state.is_0_button_held = true
            }

            context
                .coordinator_commands_sender
                .send(CoordinatorCommand::SetActivePages {
                    key_page_id: button_config.key_page.as_ref().unwrap_or(&context.state.active_key_page_id).clone(),
                    knob_page_id: button_config.knob_page.as_ref().unwrap_or(&context.state.active_knob_page_id).clone(),
                })
                .unwrap()
        }
        LoupedeckEvent::ButtonUp { button } => {
            if button == LoupedeckButton::N0 {
                context.state.is_0_button_held = false
            }
        }
        LoupedeckEvent::Touch { x, y, is_end, touch_id } => {
            let characteristics = context.device.characteristics();
            let display = characteristics.get_display_at_coordinates(x, y);

            if let Some(display) = display {
                if display == &characteristics.key_grid.display {
                    let key_index = characteristics.key_grid.get_key_at_global_coordinates(x, y);
                    if let Some(key_index) = key_index {
                        let position = KeyPosition {
                            x: (key_index % characteristics.key_grid.columns) as u16 + 1,
                            y: (key_index / characteristics.key_grid.columns) as u16 + 1,
                        };

                        let path = KeyPath {
                            page_id: context.state.active_key_page_id.clone(),
                            position,
                        };

                        let LoupedeckDisplayRect {
                            x: top_left_x, y: top_left_y, ..
                        } = characteristics
                            .key_grid
                            .get_local_key_rect(key_index)
                            .expect("the key index is valid because is was returned by get_key_at_global_coordinates");

                        let kind = if is_end {
                            context.state.active_touch_ids.remove(&touch_id);
                            KeyTouchEventKind::End
                        } else {
                            let is_new = context.state.active_touch_ids.insert(touch_id);
                            if is_new {
                                KeyTouchEventKind::Start
                            } else {
                                KeyTouchEventKind::Move
                            }
                        };

                        send_key_event(
                            path.clone(),
                            KeyEvent::Touch {
                                touch_id,
                                x: x - top_left_x,
                                y: y - top_left_y,
                                kind,
                            },
                        );

                        if kind == KeyTouchEventKind::Start {
                            send_key_event(path.clone(), KeyEvent::Press);
                        }
                    }
                }
            }
        }
        LoupedeckEvent::KnobRotate { knob, direction } => {
            if knob == LoupedeckKnob::LeftBottom && context.state.is_0_button_held {
                context.state.screen_brightness = (context.state.screen_brightness as i8)
                    .add(match direction {
                        RotationDirection::Clockwise => 1,
                        RotationDirection::Counterclockwise => -1,
                    })
                    .clamp(0, 10) as u8;

                context.device.set_brightness(context.state.screen_brightness);
            } else {
                let position: KnobPosition = get_position_of_loupedeck_knob(knob);

                send_knob_event(
                    KnobPath {
                        page_id: context.state.active_knob_page_id.clone(),
                        position,
                    },
                    KnobEvent::Rotate {
                        direction: match direction {
                            RotationDirection::Clockwise => deckster_shared::handler_communication::RotationDirection::Clockwise,
                            RotationDirection::Counterclockwise => deckster_shared::handler_communication::RotationDirection::Counterclockwise,
                        },
                    },
                )
            }
        }
        LoupedeckEvent::KnobDown { knob } => {
            let position: KnobPosition = get_position_of_loupedeck_knob(knob);

            send_knob_event(
                KnobPath {
                    page_id: context.state.active_knob_page_id.clone(),
                    position,
                },
                KnobEvent::Press,
            )
        }
        _ => {}
    }

    true
}

fn handle_coordinator_command(context: &mut IoWorkerContext, command: CoordinatorCommand) {
    log::trace!("Handling coordinator command: {:?}", &command);

    match command {
        CoordinatorCommand::SetActivePages { key_page_id, knob_page_id } => {
            context.state.active_key_page_id = key_page_id;
            context.state.active_knob_page_id = knob_page_id;

            for button in context.device.characteristics().available_buttons {
                context
                    .device
                    .set_button_color(button, get_correct_button_color(context, ButtonPosition::of(&button)))
                    .expect("the button is available for this device because that is literally what we are iterating over");
            }

            redraw_visible_page(context);
        }
        CoordinatorCommand::SetAllRemoteHostsInactive => {
            context.state.active_remote_handler_host_ids.clear();
            redraw_visible_page(context);
        }
        CoordinatorCommand::SetRemoteHostIsActive { host_id, is_active } => {
            if is_active {
                context.state.active_remote_handler_host_ids.insert(host_id);
            } else {
                context.state.active_remote_handler_host_ids.remove(&host_id);
            }

            redraw_visible_page(context);
        }
    }
}

fn handle_handler_command(context: &mut IoWorkerContext, command: HandlerCommand) {
    log::trace!("Handling handler command: {:?}", &command);

    match command {
        HandlerCommand::Vibrate { pattern } => {
            context.device.vibrate(map_vibration_pattern(pattern));
        }
        HandlerCommand::SetKeyStyle { path, value } => {
            context.state.mutate_key_for_command("SetKeyStyle", &path, |k| {
                k.style = value;
            });

            draw_key_at_path_if_visible(context, path);
            context.device.refresh_display(&context.device.characteristics().key_grid.display).unwrap();
        }
        HandlerCommand::SetKnobStyle { path, value } => {
            context.state.mutate_knob_for_command("SetKnobStyle", &path, |k| {
                k.style = value;
            });

            draw_knob_at_path_if_visible(context, path);
            context.device.refresh_display(&context.device.characteristics().key_grid.display).unwrap();
        }
        HandlerCommand::SetKnobValue { path, value } => {
            if let Some(v) = value {
                if !(0.0..=1.0).contains(&v) {
                    log::error!("Received SetKnobValue with an out-of-range value: {}", v);
                    return;
                }
            }

            context.state.mutate_knob_for_command("SetKnobValue", &path, |k| {
                k.value = value;
            });

            draw_knob_at_path_if_visible(context, path);
        }
    }
}

fn redraw_visible_page(context: &IoWorkerContext) {
    let key_grid = &context.device.characteristics().key_grid;
    for index in 0..(key_grid.rows * key_grid.columns) {
        draw_key_at_index(context, index);
    }

    for position in KnobPosition::VARIANTS {
        draw_knob_at_position(context, *position);
    }

    context.device.refresh_display(&key_grid.display).unwrap();
}

// active -> config.active_button_color
// no actions defined -> #000000
// inactive -> config.inactive_button_color
fn get_correct_button_color(context: &IoWorkerContext, button_position: ButtonPosition) -> RGB8 {
    let button_config = &context.config.buttons[button_position];

    if let Some(key_page) = &button_config.key_page {
        if key_page == &context.state.active_key_page_id {
            if let Some(knob_page) = &button_config.knob_page {
                if knob_page == &context.state.active_knob_page_id {
                    return context.config.active_button_color.into();
                }
            }
        }
    } else if button_config.knob_page.is_none() {
        return RGB8::new(0, 0, 0);
    }

    context.config.inactive_button_color.into()
}

fn get_key_index_for_position(key_grid: &LoupedeckDeviceKeyGridCharacteristics, position: KeyPosition) -> Option<u8> {
    if (position.x - 1) >= key_grid.columns as u16 || (position.y - 1) >= key_grid.rows as u16 {
        None
    } else {
        let x = (position.x - 1) as u8;
        let y = (position.y - 1) as u8;
        Some(y * key_grid.columns + x)
    }
}

fn get_key_position_for_index(key_grid: &LoupedeckDeviceKeyGridCharacteristics, index: u8) -> KeyPosition {
    let x = index % key_grid.columns;
    let y = index / key_grid.columns;

    KeyPosition {
        x: (x + 1) as u16,
        y: (y + 1) as u16,
    }
}

/// Panics when `index` is invalid.
fn draw_key(context: &IoWorkerContext, index: u8, key: Option<&Key>) {
    let key_grid = &context.device.characteristics().key_grid;
    let rect = key_grid.get_local_key_rect(index).expect("index is assumed to be valid");

    let buffer = render_key(
        &context.graphics,
        IntSize::from_wh(rect.w as u32, rect.h as u32).expect("rect.w and rect.h are not zero"),
        &context.state.active_remote_handler_host_ids,
        key,
    );
    context
        .device
        .replace_framebuffer_area_raw(&key_grid.display, rect.x, rect.y, rect.w, rect.h, buffer)
        .unwrap();
}

fn draw_key_at_index(context: &IoWorkerContext, index: u8) {
    let position = get_key_position_for_index(&context.device.characteristics().key_grid, index);

    draw_key(context, index, context.state.active_key_page().keys_by_position.get(&position));
}

fn draw_key_at_position_if_visible(context: &IoWorkerContext, position: KeyPosition) {
    let index = get_key_index_for_position(&context.device.characteristics().key_grid, position);

    if let Some(index) = index {
        draw_key(context, index, context.state.active_key_page().keys_by_position.get(&position));
    }
}

fn draw_key_at_path_if_visible(context: &IoWorkerContext, path: KeyPath) {
    if context.state.active_key_page_id == path.page_id {
        draw_key_at_position_if_visible(context, path.position);
    }
}

fn draw_knob(context: &IoWorkerContext, position: KnobPosition, knob: Option<&Knob>) {
    if let Some((display, rect)) = context.device.characteristics().get_display_and_rect_for_knob(match position {
        KnobPosition::LeftTop => LoupedeckKnob::LeftTop,
        KnobPosition::LeftMiddle => LoupedeckKnob::LeftMiddle,
        KnobPosition::LeftBottom => LoupedeckKnob::LeftBottom,
        KnobPosition::RightTop => LoupedeckKnob::RightTop,
        KnobPosition::RightMiddle => LoupedeckKnob::RightMiddle,
        KnobPosition::RightBottom => LoupedeckKnob::RightBottom,
    }) {
        let buffer = render_knob(
            &context.graphics,
            IntSize::from_wh(rect.w as u32, rect.h as u32).expect("rect.w and rect.h are not zero."),
            &context.state.active_remote_handler_host_ids,
            knob,
        );
        context
            .device
            .replace_framebuffer_area_raw(display, rect.x, rect.y, rect.w, rect.h, buffer)
            .unwrap();
    }
}

fn draw_knob_at_position(context: &IoWorkerContext, position: KnobPosition) {
    draw_knob(context, position, Some(&context.state.active_knob_page().knobs_by_position[position]));
}

fn draw_knob_at_path_if_visible(context: &IoWorkerContext, path: KnobPath) {
    if context.state.active_knob_page_id == path.page_id {
        draw_knob_at_position(context, path.position);
    }
}

fn get_position_of_loupedeck_knob(value: LoupedeckKnob) -> KnobPosition {
    match value {
        LoupedeckKnob::LeftTop => KnobPosition::LeftTop,
        LoupedeckKnob::LeftMiddle => KnobPosition::LeftMiddle,
        LoupedeckKnob::LeftBottom => KnobPosition::LeftBottom,
        LoupedeckKnob::RightTop => KnobPosition::RightTop,
        LoupedeckKnob::RightMiddle => KnobPosition::RightMiddle,
        LoupedeckKnob::RightBottom => KnobPosition::RightBottom,
    }
}

fn map_vibration_pattern(value: VibrationPattern) -> LSVibrationPattern {
    match value {
        VibrationPattern::Short => LSVibrationPattern::Short,
        VibrationPattern::Medium => LSVibrationPattern::Medium,
        VibrationPattern::Long => LSVibrationPattern::Long,
        VibrationPattern::Low => LSVibrationPattern::Low,
        VibrationPattern::ShortLow => LSVibrationPattern::ShortLow,
        VibrationPattern::ShortLower => LSVibrationPattern::ShortLower,
        VibrationPattern::Lower => LSVibrationPattern::Lower,
        VibrationPattern::Lowest => LSVibrationPattern::Lowest,
        VibrationPattern::DescendSlow => LSVibrationPattern::DescendSlow,
        VibrationPattern::DescendMed => LSVibrationPattern::DescendMed,
        VibrationPattern::DescendFast => LSVibrationPattern::DescendFast,
        VibrationPattern::AscendSlow => LSVibrationPattern::AscendSlow,
        VibrationPattern::AscendMed => LSVibrationPattern::AscendMed,
        VibrationPattern::AscendFast => LSVibrationPattern::AscendFast,
        VibrationPattern::RevSlowest => LSVibrationPattern::RevSlowest,
        VibrationPattern::RevSlow => LSVibrationPattern::RevSlow,
        VibrationPattern::RevMed => LSVibrationPattern::RevMed,
        VibrationPattern::RevFast => LSVibrationPattern::RevFast,
        VibrationPattern::RevFaster => LSVibrationPattern::RevFaster,
        VibrationPattern::RevFastest => LSVibrationPattern::RevFastest,
        VibrationPattern::RiseFall => LSVibrationPattern::RiseFall,
        VibrationPattern::Buzz => LSVibrationPattern::Buzz,
        VibrationPattern::Rumble5 => LSVibrationPattern::Rumble5,
        VibrationPattern::Rumble4 => LSVibrationPattern::Rumble4,
        VibrationPattern::Rumble3 => LSVibrationPattern::Rumble3,
        VibrationPattern::Rumble2 => LSVibrationPattern::Rumble2,
        VibrationPattern::Rumble1 => LSVibrationPattern::Rumble1,
        VibrationPattern::VeryLong => LSVibrationPattern::VeryLong,
    }
}
